#!/bin/python3

import abc

from enum import Enum


class UpdateContext:
    
    def __init__(self, config, manager):
        self._config = config
        self._manager = manager

    @property
    def config(self):
        return self._config

    def write_config_changes(self):
        self._manager.write(self._config)

class Operation(str, Enum):
    INSERT = 'Insert'
    DELETE = 'Delete'
    UPDATE = 'Update'
    UPSERT = 'Upsert'

class IdempotentUpdateStep(metaclass=abc.ABCMeta):

    @property
    @abc.abstractmethod
    def operation(self):
        return

    @property
    @abc.abstractmethod
    def description(self):
        return
    
    @abc.abstractmethod
    def run(self, context):
        return

class AddServiceMemoryAllocationStep(IdempotentUpdateStep):

    def __init__(self, service_name, allocation_percent, restart_service=True):
        self._service_name = service_name
        self._allocation_percent = allocation_percent
        self._restart_service = restart_service

    @property
    def operation(self):
        return Operation.INSERT

    @property
    def description(self):
        return 'Adds service memory allocation for "{}". The default value is {}%.'.format(self._service_name, self._allocation_percent)

    def run(self, context):
        context.config.memory.allocation.add_service_if_not_exists(self._service_name, self._allocation_percent, restart_service=self._restart_service)


class Updater:

    def __init__(self, config, manager):
        self._config = config
        self._manager = manager
        self._steps = [
            AddServiceMemoryAllocationStep('reverse-proxy', 5.0)
        ]

    @property
    def steps(self):
        return self._steps

    def update(self):

        context = UpdateContext(self._config, self._manager)

        for step in self._steps:
            step.run(context)

        context.write_config_changes()
