import click
import humanize

from tabulate import tabulate

from cli.format import indent, title, subtitle, subsubtitle, subsubsubtitle
from cli.session import obtain_session
from log.collection.collection import LogCollectionService

log_collow_show_headers = [
    '\nID',
    '\nType',
    'Steps\nComplete',
    '\n%',
    '\nExecution Time',
    '\nCompleted',
    '\nSize',
    '\nDestination'
]


@click.group()
def log():
    pass

@log.group()
def collection():
    pass

@collection.group()
def collect():
    """
    Creates a new log collection.
    """
    pass

@collect.command('system')
@click.pass_context
@click.option('--username', default=None, help='SPP username')
@click.option('--password', default=None, help='SPP password')
@click.option('--token', help='TOTP token for the SPP user')
def log_collect_system(ctx, username, password, token):
    """
    Starts a system log collection.
    """

    try:
        session = obtain_session(username, password, token)
        log_service = LogCollectionService(session)

        log_collection = log_service.start_system_log_collection()

        click.echo('')
        title('System Log Collection')
        click.echo('')
        click.echo('  Successfully started system log collection')
        click.echo('')
        subtitle('Information')
        click.echo('')
        click.echo('  ID:          {}'.format(log_collection.id))
        click.echo('  Destination: {}'.format(log_collection.destination))
        click.echo('')
        subtitle('Progress')
        click.echo('')
        click.echo('  Progress for this log collection can be tracked using the following command:')
        click.echo('')
        click.echo('    spp log collection get {}'.format(log_collection.id))
        click.echo('')

    except Exception as e:

        click.echo('')
        click.echo('  Error: {}'.format(str(e)))
        click.echo('')

@collection.command('show')
@click.pass_context
@click.option('--username', default=None, help='SPP username')
@click.option('--password', default=None, help='SPP password')
@click.option('--token', help='TOTP token for the SPP user')
def show_log_collections(ctx, username, password, token):
    """
    Shows the information for all available log collections.
    """

    try:
        session = obtain_session(username, password, token)
        log_service = LogCollectionService(session)

        collections = log_service.find_collections()

        click.echo('')
        title('Log Collections')
        click.echo('')
        print_collections(collections)
        click.echo('')

    except Exception as e:

        click.echo('')
        click.echo('  Error: {}'.format(str(e)))
        click.echo('')

        raise e

def print_collections(collections, indent_spaces=2):

    rows = format_log_collection_rows(collections)

    if len(rows) > 0:
        click.echo(indent(tabulate(rows, headers=log_collow_show_headers), space=indent_spaces))
    else:
        click.echo(indent('None', space=indent_spaces))

def format_log_collection_rows(collections):

    rows = []

    for c in collections.collections:

        steps = '{} / {}'.format(c.completed_step_count, c.total_step_count)
        percent = int(c.percent_complete)
        execution_time = format_time(c.total_execution_times.total)
        size = humanize.naturalsize(c.size)

        if c.completion:
            completed_ago = humanize.naturaltime(c.completed_ago)
        elif c.state == 'STARTED':
            completed_ago = '(Still in progress)'
        elif c.state == 'ABORTED':
            completed_ago = '(Aborted)'
        else:
            completed_ago = '(Unknown)'

        rows.append((c.id, c.type_display_name, steps, percent, execution_time, completed_ago, size, c.destination))

    return rows

def format_time(millis):
    
    total_seconds = int(millis / 1000)
    hours = int(total_seconds / 3600)
    minutes = int((total_seconds - (hours * 3600)) / 60)
    seconds = int(total_seconds - (minutes * 60) - (hours * 3600))
    remaining_millis = millis - (seconds * 1000) - (minutes * 60 * 1000) - (hours * 3600 * 1000)

    if hours > 0:
        return '{}h {}m {}s'.format(hours, minutes, seconds)
    elif minutes > 0:
        return '{}m {}s'.format(minutes, seconds)
    elif seconds > 0:
        return '{}s {}ms'.format(seconds, remaining_millis)
    else:
        return '{}ms'.format(remaining_millis)


@collection.command('get')
@click.pass_context
@click.argument('id')
@click.option('--username', default=None, help='SPP username')
@click.option('--password', default=None, help='SPP password')
@click.option('--token', help='TOTP token for the SPP user')
@click.option('--details', '-d', is_flag=True, help='Show all details for the log collection, such as step execution details')
def show_log_collections(ctx, id, username, password, token, details):
    """
    Shows the information for the log collection corresponding to the supplied ID.
    """

    try:
        session = obtain_session(username, password, token)
        log_service = LogCollectionService(session)

        collection = log_service.find_collection_by_id(id)

        if not collection:

            click.echo('')
            click.echo('  No log collection found with the ID {}'.format(id))
            click.echo('')
        
            return

        steps_headers = [
            'State',
            'Count'
        ]

        execution_time_headers = [
            'Event',
            'Time',
            'Time (ms)'
        ]

        step_executions_headers = [
            'Step',
            'State',
            'Time',
            'Time (ms)'
        ]

        steps_rows = [
            ('In Progress', collection.steps.summary.states.in_progress),
            ('Completed', collection.steps.summary.states.completed),
            ('Successful', collection.steps.summary.states.successful),
            ('Failed', collection.steps.summary.states.failed),
            ('Timed Out', collection.steps.summary.states.timed_out),
            ('Skipped', collection.steps.summary.states.skipped)
        ]

        total_execution_time_rows = [
            create_time_row('Queuing', collection.total_execution_times.queuing),
            create_time_row('Waiting to Start', collection.total_execution_times.waiting_to_start),
            create_time_row('Running', collection.total_execution_times.running),
            create_time_row('Total', collection.total_execution_times.total)
        ]

        cumulative_execution_time_rows = [
            create_time_row('Queued', collection.steps.summary.cumulative_execution_times.queued),
            create_time_row('Collecting', collection.steps.summary.cumulative_execution_times.collecting),
            create_time_row('Writing', collection.steps.summary.cumulative_execution_times.writing),
            create_time_row('Cleaning Up', collection.steps.summary.cumulative_execution_times.cleaning_up),
            create_time_row('Idle', collection.steps.summary.cumulative_execution_times.idle),
            create_time_row('Waiting to Collect', collection.steps.summary.cumulative_execution_times.waiting_to_collect),
            create_time_row('Waiting to Write', collection.steps.summary.cumulative_execution_times.waiting_to_write),
            create_time_row('Waiting to Cleanup', collection.steps.summary.cumulative_execution_times.waiting_to_cleanup),
            create_time_row('Pending Completion', collection.steps.summary.cumulative_execution_times.pending_completion),
            create_time_row('Total', collection.steps.summary.cumulative_execution_times.total),
        ]
        
        click.echo('')
        title('Log Collection')
        click.echo('')
        click.echo('  ID:          {}'.format(collection.id))
        click.echo('  Type:        {}'.format(collection.type_display_name))
        click.echo('  State:       {}'.format(collection.state_display_name))
        click.echo('  Destination: {}'.format(collection.destination))
        click.echo('  Exists:      {}'.format(collection.exists))
        click.echo('  Size:        {}'.format(humanize.naturalsize(collection.size)))

        if collection.completion:
            click.echo('  Completed:   {} ({})'.format(humanize.naturaltime(collection.completed_ago), collection.completion))
        else:
            click.echo('  Completed:   (Still in progress)')
            
        click.echo('')
        subtitle('Total Execution Times')
        click.echo('')
        click.echo(indent(tabulate(total_execution_time_rows, headers=execution_time_headers), space=2))
        click.echo('')
        subtitle('Cumulative Execution Times')
        click.echo('')
        click.echo(indent(tabulate(cumulative_execution_time_rows, headers=execution_time_headers), space=2))
        click.echo('')
        subtitle('Steps ({})'.format(len(collection.steps.executions)))
        click.echo('')
        click.echo('  Total:            {}'.format(collection.steps.summary.total))
        click.echo('  Completed:        {}'.format(collection.steps.summary.states.completed))
        click.echo('  Percent Complete: {}%'.format(int(collection.percent_complete)))
        click.echo('')
        click.echo(indent(tabulate(steps_rows, headers=steps_headers), space=2))
        click.echo('')
        click.echo(indent(tabulate(create_step_executions_rows(collection), headers=step_executions_headers), space=2))
        click.echo('')

        if details:

            timestamps_headers = [
                'Event',
                'Timestamp'
            ]

            outcomes_headers = [
                'Event',
                'Outcome'
            ]

            click.echo('')
            subsubtitle('Executions')
            click.echo('')

            for se in collection.steps.executions:

                se_execution_time_rows = [
                    create_time_row('Queued', se.execution_times.queued),
                    create_time_row('Collecting', se.execution_times.collecting),
                    create_time_row('Writing', se.execution_times.writing),
                    create_time_row('Cleaning Up', se.execution_times.cleaning_up),
                    create_time_row('Idle', se.execution_times.idle),
                    create_time_row('Waiting to Collect', se.execution_times.waiting_to_collect),
                    create_time_row('Waiting to Write', se.execution_times.waiting_to_write),
                    create_time_row('Waiting to Cleanup', se.execution_times.waiting_to_cleanup),
                    create_time_row('Pending Completion', se.execution_times.pending_completion),
                    create_time_row('Total', se.execution_times.total)
                ]

                se_outcomes_rows = [
                    ('Collecting', se.outcomes.collecting_display_name),
                    ('Writing', se.outcomes.writing_display_name),
                    ('Cleanup', se.outcomes.cleanup_display_name)
                ]

                se_timestamps_rows = [
                    ('Queued', se.timestamps.queued),
                    ('Started', se.timestamps.started),
                    ('Collecting Started', se.timestamps.collecting_started),
                    ('Colelcting Completed', se.timestamps.collecting_completed),
                    ('Writing Started', se.timestamps.writing_started),
                    ('Writing Completed', se.timestamps.writing_completed),
                    ('Cleanup Started', se.timestamps.cleanup_started),
                    ('Cleanup Completed', se.timestamps.cleanup_completed),
                    ('Completed', se.timestamps.completed)
                ]

                click.echo('')
                subsubsubtitle('{}'.format(se.step_name))
                click.echo('')
                click.echo('  State: {}'.format(se.current_state_display_name))
                click.echo('')
                click.echo('  Execution Times:')
                click.echo('')
                click.echo(indent(tabulate(se_execution_time_rows, headers=execution_time_headers), space=4))
                click.echo('')
                click.echo('  Outcomes:')
                click.echo('')
                click.echo(indent(tabulate(se_outcomes_rows, headers=outcomes_headers), space=4))
                click.echo('')
                click.echo('  Timestamps:')
                click.echo('')
                click.echo(indent(tabulate(se_timestamps_rows, headers=timestamps_headers), space=4))
                click.echo('')

    except Exception as e:

        click.echo('')
        click.echo('  Error: {}'.format(str(e)))
        click.echo('')

def create_time_row(title, millis):
    return (title, format_time(millis), millis)

def create_step_executions_rows(collection):

    rows = []

    for se in collection.steps.executions:

        rows.append((se.step_name, se.current_state_display_name, format_time(se.execution_times.total), se.execution_times.total))

    return rows


@collection.command('expire')
@click.pass_context
@click.option('--username', default=None, help='SPP username')
@click.option('--password', default=None, help='SPP password')
@click.option('--token', help='TOTP token for the SPP user')
def expire_log_collections(ctx, username, password, token):
    """
    Removes all expired log collections.
    """

    try:
        session = obtain_session(username, password, token)
        log_service = LogCollectionService(session)

        log_service.expire()

        collections = log_service.find_collections()       

        click.echo('')
        title('Expire Log Collections')
        click.echo('')
        click.echo('  Expired log collections have been removed')
        click.echo('')
        subtitle('Remaining Log Collections')
        click.echo('')
        print_collections(collections)
        click.echo('')

    except Exception as e:
        click.echo('')
        click.echo('  Error: {}'.format(str(e)))
        click.echo('')


@collection.command('delete')
@click.argument('id')
@click.pass_context
@click.option('--username', default=None, help='SPP username')
@click.option('--password', default=None, help='SPP password')
@click.option('--token', help='TOTP token for the SPP user')
def delete_log_collection_by_id(ctx, id, username, password, token):
    """
    Deletes the log collection associated with the supplied ID.
    """

    try:
        session = obtain_session(username, password, token)
        log_service = LogCollectionService(session)

        log_service.delete_by_id(id)

        collections = log_service.find_collections()       

        click.echo('')
        title('Delete Log Collection')
        click.echo('')
        click.echo('  Deleted log collection with ID {}'.format(id))
        click.echo('')
        subtitle('Remaining Log Collections')
        click.echo('')
        print_collections(collections)
        click.echo('')

    except Exception as e:
        click.echo('')
        click.echo('  Error: {}'.format(str(e)))
        click.echo('')


@collection.command('delete-all')
@click.pass_context
@click.option('--username', default=None, help='SPP username')
@click.option('--password', default=None, help='SPP password')
@click.option('--token', help='TOTP token for the SPP user')
def delete_all_log_collections(ctx, username, password, token):
    """
    Deletes all log collections.
    """

    try:
        session = obtain_session(username, password, token)
        log_service = LogCollectionService(session)

        log_service.delete_all()       

        click.echo('')
        title('Delete Log Collection')
        click.echo('')
        click.echo('  Deleted all log collections')
        click.echo('')

    except Exception as e:
        click.echo('')
        click.echo('  Error: {}'.format(str(e)))
        click.echo('')

