import datetime as dt


class LogCollectionService:

    def __init__(self, session):
        self._session = session

    def start_system_log_collection(self):

        response = self._session.post('http://localhost:8082/api/log/system?action=collect').json()

        return LogCollection(response)

    def find_collections(self):

        response = self._session.get('http://localhost:8082/api/log/collection').json()

        return LogCollections(response)

    def find_collection_by_id(self, id):

        response = self._session.get('http://localhost:8082/api/log/collection/{}'.format(id))

        if response.status_code == 404:
            return None
        else:
            return LogCollection(response.json())

    def expire(self):
        return self._session.post('http://localhost:8082/api/log/collection?action=expire')

    def delete_by_id(self, id):
        return self._session.delete('http://localhost:8082/api/log/collection/{}'.format(id))

    def delete_all(self):
        return self._session.delete('http://localhost:8082/api/log/collection')

class LogCollections:

    def __init__(self, json):
        self._from_json(json)

    def _from_json(self, json):
        self.collections = map(lambda c: LogCollection(c), json['collections'])


class LogCollection:

    def __init__(self, json):
        self._from_json(json)

    def _from_json(self, json):
        self.id = json['id']
        self.type = json['type']
        self.type_display_name = json['typeDisplayName']
        self.destination = json['destination']
        self.start = dt.datetime.utcfromtimestamp(json['startTimestamp'] / 1000.0)
        self.completion = self._find_completion(json)
        self.state = json['state']
        self.state_display_name = json['stateDisplayName']
        self.exists = bool(json['exists'])
        self.size = int(json['sizeInBytes'])
        self.total_execution_times = TotalExecutionTimes(json['totalExecutionTimes'])
        self.steps = Steps(json['steps'])

    def _find_completion(self, json):

        if 'completionTimestamp' in json and json['completionTimestamp']:
            return dt.datetime.utcfromtimestamp(json['completionTimestamp'] / 1000.0)
        else:
            return None

    @property
    def completed_step_count(self):
        return self.steps.summary.states.completed

    @property
    def total_step_count(self):
        return self.steps.summary.total

    @property
    def percent_complete(self):

        if self.total_step_count == 0:
            return 100
        else:
            return (self.completed_step_count / self.total_step_count) * 100

    @property
    def completed_ago(self):

        if self.completion:
            return dt.datetime.utcnow() - self.completion
        else:
            return None


class TotalExecutionTimes:

    def __init__(self, json):
        self._from_json(json)

    def _from_json(self, json):
        self.total = json['total']
        self.queuing = json['queuing']
        self.waiting_to_start = json['waitingToStart']
        self.running = json['running']


class Steps:

    def __init__(self, json):
        self._from_json(json)

    def _from_json(self, json):
        self.executions = list(map(lambda e: StepExecution(e), json['executions']))
        self.summary = StepsSummary(json['summary'])


class StepExecution:
    
    def __init__(self, json):
        self._from_json(json)

    def _from_json(self, json):
        self.step_name = json['stepName']
        self.current_state = json['currentState']
        self.current_state_display_name = json['currentStateDisplayName']
        self.execution_times = ExecutionTimes(json['executionTimes'])
        self.outcomes = StepOutcomes(json['outcomes'])
        self.timestamps = StepTimestamps(json['timestamps'])


class StepOutcomes:
    
    def __init__(self, json):
        self._from_json(json)

    def _from_json(self, json):
        self.collecting = json['collecting']
        self.writing = json['writing']
        self.cleanup = json['cleanup']
        self.collecting_display_name = json['collectingDisplayName']
        self.writing_display_name = json['writingDisplayName']
        self.cleanup_display_name = json['cleanupDisplayName']


class StepTimestamps:

    def __init__(self, json):
        self._from_json(json)

    def _from_json(self, json):
        self.queued = json['queued']
        self.started = json['started']
        self.collecting_started = json['collectingStarted']
        self.collecting_completed = json['collectingCompleted']
        self.writing_started = json['writingStarted']
        self.writing_completed = json['writingCompleted']
        self.cleanup_started = json['cleanupStarted']
        self.cleanup_completed = json['cleanupCompleted']
        self.completed = json['completed']


class StepsSummary:

    def __init__(self, json):
        self._from_json(json)

    def _from_json(self, json):
        self.total = json['total']
        self.states = StepStates(json['states'])
        self.cumulative_execution_times = ExecutionTimes(json['cumulativeExecutonTimes'])


class StepStates:

    def __init__(self, json):
        self._from_json(json)

    def _from_json(self, json):
        self.in_progress = json['inProgress']
        self.completed = json['completed']
        self.successful = json['successful']
        self.failed = json['failed']
        self.timed_out = json['timedOut']
        self.skipped = json['skipped']


class ExecutionTimes:

    def __init__(self, json):
        self._from_json(json)

    def _from_json(self, json):
        self.total = json['total']
        self.queued = json['queued']
        self.collecting = json['collecting']
        self.writing = json['writing']
        self.cleaning_up = json['cleaningUp']
        self.idle = json['idle']
        self.waiting_to_collect = json['waitingToCollect']
        self.waiting_to_write = json['waitingToWrite']
        self.waiting_to_cleanup = json['waitingToCleanup']
        self.pending_completion = json['pendingCompletion']