import click

from tabulate import tabulate

from cli.format import display_units, title, indent
from config.system import ConfigurationManager
from memory.memory import SystemMemory, ByteValue
from memory.service import ServiceFinder
from memory.error import NegativeAllocationPercentException, ReservedAllocationWouldBeOver100PercentException


@click.group()
def memory():
    pass


@memory.group()
def system():
    pass


@memory.group()
def allocation():
    pass


@system.command('show')
@click.option('--units', type=click.Choice(['b', 'kb', 'mb', 'gb'], case_sensitive=False), default='gb', help='Units to display')
@click.option('--precision', type=int, default=2, help='Decimal places to display')
def show_system_memory(units, precision):

    sys_mem = SystemMemory()
    
    headers = ['Type', 'Value ({})'.format(units.upper()), '%']
    rows = [
        ('Total', display_units(sys_mem.total, units), 100),
        ('Available', display_units(sys_mem.available, units), sys_mem.available.percent),
        ('Used', display_units(sys_mem.used, units), sys_mem.used.percent),
        ('Free', display_units(sys_mem.available, units), sys_mem.free.percent),
        ('Active', display_units(sys_mem.active, units), sys_mem.active.percent),
        ('Inactive', display_units(sys_mem.inactive, units), sys_mem.inactive.percent),
        ('Buffers', display_units(sys_mem.buffers, units), sys_mem.buffers.percent),
        ('Cached', display_units(sys_mem.cached, units), sys_mem.cached.percent),     
        ('Shared', display_units(sys_mem.shared, units), sys_mem.shared.percent),
        ('Slab', display_units(sys_mem.slab, units), sys_mem.slab.percent),   
    ]

    click.echo('')
    title('System Memory')
    click.echo('')
    click.echo(indent(tabulate(rows, headers=headers, floatfmt=".{}f".format(precision)), space=2))
    click.echo('')


@allocation.command('show', help='Shows all memory allocation information')
@click.pass_context
@click.option('--units', type=click.Choice(['b', 'kb', 'mb', 'gb'], case_sensitive=False), default='gb', help='Units to display')
@click.option('--precision', type=int, default=2, help='Decimal places to display')
def show_memory_allocations(ctx, units, precision):
    
    sys_mem = SystemMemory()
    total_mem = sys_mem.total

    config_manager = ConfigurationManager(ctx.obj['config_path'])
    config = config_manager.load()

    service_finder = ServiceFinder(config, sys_mem)

    services = service_finder.find_all()

    total_allocation_bytes = sum([u.allocation.bytes for u in services])
    total_allocation = ByteValue(total_allocation_bytes, total_mem.bytes)
    unreserved = ByteValue(total_mem.bytes - total_allocation.bytes, total_mem.bytes)

    headers = ['Service', 'Memory ({})'.format(units.upper()), '%']
    rows = []

    reserved_headers = ['Category', 'Memory ({})'.format(units.upper()), '%']
    reserved_rows = [
        ('Allocated', display_units(total_allocation, units), total_allocation.percent),
        ('Unallocated', display_units(unreserved, units), unreserved.percent)
    ]

    for service in services:
        row = (service.name, display_units(service.allocation, units), service.allocation_percent)
        rows.append(row)
    
    click.echo('')
    title('Services')
    click.echo('')
    click.echo(indent(tabulate(rows, headers=headers, floatfmt=".{}f".format(precision)), space=2))
    click.echo('')
    title('Allocations')
    click.echo('')
    click.echo(indent(tabulate(reserved_rows, headers=reserved_headers, floatfmt=".{}f".format(precision)), space=2))
    click.echo('')
    

@allocation.command('get', help='Shows memory allocation information for a specific service')
@click.pass_context
@click.option('--service', 'service_name', required=True, help='The service whose memory allocation to obtain')
@click.option('--units', type=click.Choice(['b', 'kb', 'mb', 'gb', 'percent'], case_sensitive=False), default='gb', help='Units to display')
@click.option('--precision', type=int, default=2, help='Decimal places to display')
def get_memory_allocation(ctx, service_name, units, precision):
    
    sys_mem = SystemMemory()

    config_manager = ConfigurationManager(ctx.obj['config_path'])
    config = config_manager.load()

    service_finder = ServiceFinder(config, sys_mem)

    service = service_finder.find_by_name(service_name)

    if service is not None:
        if units == 'percent':
            click.echo('{:.{prec}f}'.format(service.allocation_percent, prec=precision))
        else:
            allocation = display_units(service.allocation, units)
            click.echo('{:.{prec}f}'.format(allocation, prec=precision))

    else:
        error('Cannot find service "{}"'.format(service_name))


def error(message):
    click.echo(message)
    exit(1)


@allocation.command('set', help='Configures memory allocation information for a specific service')
@click.pass_context
@click.option('--service', 'service_name', required=True, help='The service whose memory allocation to set')
@click.option('--percent', required=True, type=float, help='The percent of memory allocated to the service, as a decimal value')
@click.option('--restart', is_flag=True, help="Whether to restart the service after setting the memory allocation")
def set_memory_allocation(ctx, service_name, percent, restart):
    
    config_manager = ConfigurationManager(ctx.obj['config_path'])
    config = config_manager.load()
    
    sys_mem = SystemMemory()
    service_finder = ServiceFinder(config, sys_mem)

    services = service_finder.find_all()
    reserved_percent = sum([s.allocation_percent for s in services])

    try:
        config_manager = ConfigurationManager(ctx.obj['config_path'])
        config = config_manager.load()

        config.memory.allocation.update_or_add_service(service_name, percent, restart)

        config_manager.write(config)

        click.echo('Set allocation for service "{}" to {}%'.format(service_name, percent))

    except PermissionError as e:
        error(f'Failed to update configuration for service {service_name} due to permission issues. Try running this command using sudo.')

    except NegativeAllocationPercentException as e:
        error('Cannot set allocation for service "{}" to less than 0%'.format(service_name))

    except ReservedAllocationWouldBeOver100PercentException as e:
        error('Cannot set allocation for service "{}" to {}%. Setting this value would set total allocations to over 100%'.format(service_name, percent))


@allocation.command('unset', help='Removes memory allocation information for a specific service')
@click.pass_context
@click.option('--service', 'service_name', required=True, help='The service whose memory allocation to unset')
def unset_memory_allocation(ctx, service_name):
    
    config_manager = ConfigurationManager(ctx.obj['config_path'])
    config = config_manager.load()

    sys_mem = SystemMemory()
    service_finder = ServiceFinder(config, sys_mem)

    services = service_finder.find_all()
    reserved_percent = sum([s.allocation_percent for s in services])

    config_manager = ConfigurationManager(ctx.obj['config_path'])
    config = config_manager.load()

    config.memory.allocation.remove_service_by_name(service_name)

    config_manager.write(config)

    click.echo('Removed allocation for service "{}"'.format(service_name))