import click

from tabulate import tabulate

from cli.session import obtain_session
from cli.format import title, indent, subtitle
from references.references import ReferenceManager


@click.group()
def references():
    pass


@references.group()
def vertex():
    pass


def format_vertex_name(name, name_length):

    if name is None:
        return '<unnamed>'

    elif len(name) > name_length:
        return name[:name_length] + '...'

    else:
        return name


@vertex.command('get')
@click.pass_context
@click.argument('vertex')
@click.option('--username', default=None, help='SPP username')
@click.option('--password', default=None, help='SPP password')
@click.option('--token', help='TOTP token for the SPP user')
@click.option('--name-length', default=80, help='The maximum character length of the vertex name')
def get_vertex(ctx, vertex, username, password, token, name_length):
    """
    Obtains the information for a specific vertex.

    VERTEX is formatted as <type>:<resource-id>. For example: storageprofile:1000
    """

    try:
        session = obtain_session(username, password, token)
        (type, resource_id) = parse_vertex_name(vertex)

        ref_manager = ReferenceManager(session)
        vertex = ref_manager.find_vertex(type, resource_id)
        edges = ref_manager.find_all_edges(type, resource_id)

        incoming_edges_headers = [
            'Label\n\n', 
            'Outgoing Vertex\n\nName',
            '\n\nID',
            '\n\nType',
            '\n\nResource ID'
        ]
        incoming_edges_rows = format_edge_rows(edges.incoming, name_length, lambda edge:edge.from_vertex)

        outgoing_edges_headers = [
            'Label\n\n', 
            'Incoming Vertex\n\nName',
            '\n\nID',
            '\n\nType',
            '\n\nResource ID'
        ]
        outgoing_edges_rows = format_edge_rows(edges.outgoing, name_length, lambda edge:edge.to_vertex)

        click.echo('')
        title('Vertex Information')
        click.echo('')
        click.echo('  ID:          {}'.format(vertex.id))
        click.echo('  Name:        {}'.format(vertex.name))
        click.echo('  Type:        {}'.format(vertex.type))
        click.echo('  Resource ID: {}'.format(vertex.resource_id))
        click.echo('')
        title('Edges')
        click.echo('')
        subtitle('Incoming Edges')
        click.echo('')

        if len(incoming_edges_rows) > 0:
            click.echo(indent(tabulate(incoming_edges_rows, headers=incoming_edges_headers), space=2))
        else:
            click.echo("None")
        
        click.echo('')
        click.echo('')
        subtitle('Outgoing Edges')
        click.echo('')

        if len(outgoing_edges_rows) > 0:
            click.echo(indent(tabulate(outgoing_edges_rows, headers=outgoing_edges_headers), space=2))
        else:
            click.echo("None")
        
        click.echo('')

    except Exception as e:

        click.echo('')
        click.echo('  Error: {}'.format(str(e)))
        click.echo('')


def parse_vertex_name(name):
    
    split_name = name.split(':')
    
    type = split_name[0]
    resource_id = split_name[1]

    return (type, resource_id)


def format_edge_rows(edges, name_length, edge_direction_extractor):

    rows = []

    for edge in edges:

        vertex = edge_direction_extractor(edge)
        vertex_name = (vertex.name[:name_length] + '...') if len(vertex.name) > name_length else vertex.name
        
        row = (edge.label, vertex_name, vertex.id, vertex.type, vertex.resource_id)
        
        rows.append(row)

    return rows


@vertex.command('delete')
@click.pass_context
@click.argument('vertex')
@click.option('--username', default=None, help='SPP username')
@click.option('--password', default=None, help='SPP password')
@click.option('--token', help='TOTP token for the SPP user')
@click.option('--force', '-f', is_flag=True, help="Forces the command the execute without confirmation")
def delete_reference_vertex(ctx, vertex, username, password, token, force):
    """
    Deletes a specific vertex.

    VERTEX is formatted as <type>:<resource-id>. For example: storageprofile:1000
    """

    try:
        confirm(force, 'Vertex not deleted.')

        session = obtain_session(username, password, token)
        (type, resource_id) = parse_vertex_name(vertex)

        ref_manager = ReferenceManager(session)
        edges = ref_manager.delete_vertex(type, resource_id)

        click.echo('')
        click.echo('  Successfully deleted vertex {}.'.format(vertex))
        click.echo('')

    except Exception as e:

        click.echo('')
        click.echo('  Error: {}'.format(str(e)))
        click.echo('')


def confirm(force, message):

    if not force:

        click.echo('')
        confirmation = input('  Are you sure you want to perform this action? [Y/n] ')

        if not (confirmation == 'Y' or confirmation == 'y'):
            click.echo('')
            click.echo('  {}'.format(message))
            click.echo('')
            exit(0)


@references.group()
def edge():
    pass


@edge.group('in')
def edge_in():
    pass

@edge_in.command('delete')
@click.pass_context
@click.argument('vertex')
@click.option('--username', default=None, help='SPP username')
@click.option('--password', default=None, help='SPP password')
@click.option('--token', help='TOTP token for the SPP user')
@click.option('--label', default="USES", help='The label of the edges to delete')
@click.option('--force', '-f', is_flag=True, help="Forces the command the execute without confirmation")
def delete_vertex_edges_in(ctx, vertex, username, password, token, label, force):
    """
    Deletes the incoming edges for a specific vertex.

    VERTEX is formatted as <type>:<resource-id>. For example: storageprofile:1000
    """

    try:
        confirm(force, 'Edges not deleted.')

        session = obtain_session(username, password, token)
        (type, resource_id) = parse_vertex_name(vertex)

        ref_manager = ReferenceManager(session)
        edges = ref_manager.delete_incoming_edges(type, resource_id, label)

        click.echo('')
        click.echo('  Successfully deleted incoming edges with label {} for vertex {}.'.format(label, vertex))
        click.echo('')

    except Exception as e:

        click.echo('')
        click.echo('  Error: {}'.format(str(e)))
        click.echo('')


@vertex.command('show', help='Shows all reference vertices')
@click.pass_context
@click.option('--username', help='SPP username')
@click.option('--password', help='SPP password')
@click.option('--token', help='TOTP token for the SPP user')
@click.option('--type', default=None, help='The type of the vertex to filter by')
@click.option('--name-length', default=80, help='The maximum character length of the vertex name')
def show_vertices(ctx, username, password, token, type, name_length):

    try:
        session = obtain_session(username, password, token)

        ref_manager = ReferenceManager(session)
        vertices = ref_manager.find_all_vertices()

        headers = ['Name', 'ID', 'Type', 'Resource ID']
        rows = []

        if type is not None:
            vertices = filter(lambda v: v.type == type, vertices)

        for vertex in vertices:

            name = format_vertex_name(vertex.name, name_length)
            
            row = (name, vertex.id, vertex.type, vertex.resource_id)
            
            rows.append(row)
        
        click.echo('')
        title('Vertices')
        click.echo('')
        click.echo(indent(tabulate(rows, headers=headers), space=2))
        click.echo('')

    except Exception as e:

        click.echo('')
        click.echo('  Error: {}'.format(str(e)))
        click.echo('')
    
    
@edge_in.command('get')
@click.pass_context
@click.argument('vertex')
@click.option('--username', default=None, help='SPP username')
@click.option('--password', default=None, help='SPP password')
@click.option('--token', help='TOTP token for the SPP user')
@click.option('--name-length', default=80, help='The maximum character length of the vertex name')
def get_incoming_edges(ctx, vertex, username, password, token, name_length):
    """
    Obtains the information about incoming edges for a vertex.

    VERTEX is formatted as <type>:<resource-id>. For example: storageprofile:1000
    """

    try:
        session = obtain_session(username, password, token)
        (type, resource_id) = parse_vertex_name(vertex)

        ref_manager = ReferenceManager(session)
        edges = ref_manager.find_all_edges(type, resource_id)

        incoming_edges_headers = [
            'Label\n\n', 
            'Outgoing Vertex\n\nName',
            '\n\nID',
            '\n\nType',
            '\n\nResource ID'
        ]
        incoming_edges_rows = format_edge_rows(edges.incoming, name_length, lambda edge:edge.from_vertex)
        
        click.echo('')
        title('Incoming Edges')
        click.echo('')

        if len(incoming_edges_rows) > 0:
            click.echo(indent(tabulate(incoming_edges_rows, headers=incoming_edges_headers), space=2))
        else:
            click.echo("None")
        
        click.echo('')

    except Exception as e:

        click.echo('')
        click.echo('  Error: {}'.format(str(e)))
        click.echo('')